class
	ANTLR_TRANSFORMATION_UTILITIES

inherit
	ANTLR_AST_TYPE_CONSTANTS

feature {ANY}

	process_statement (node: ANTLR_AST_NODE; parent: AST_NODE): AST_STATEMENT
		require
			node /= Void
		do
			inspect
				node.type
			when Ast_statement then
				Result := process_statement (node.children.first, parent)
			when Ast_compound_statement then
				Result := create {AST_COMPOUND_STATEMENT}.make_from_antlr (node, parent)
			when Ast_jump_statement then
				Result := create {AST_JUMP_STATEMENT}.make_from_antlr (node, parent)
			when Ast_expression_statement then
				Result := create {AST_EXPRESSION_STATEMENT}.make_from_antlr (node, parent)
			when Ast_if_statement then
				Result := create {AST_IF_STATEMENT}.make_from_antlr (node, parent)
			when C_kw_switch then
				Result := create {AST_SWITCH_STATEMENT}.make_from_antlr (node, parent)
			when Ast_while_statement then
				Result := create {AST_WHILE_STATEMENT}.make_from_antlr (node, parent)
			when Ast_labeled_statement then
				Result := create {AST_LABELED_STATEMENT}.make_from_antlr (node, parent)
			when Ast_assembler_statement then
				Result := create {AST_ASSEMBLER_STATEMENT}.make_from_antlr (node, parent)
			when C_kw_case then
				Result := create {AST_SWITCH_CASE}.make_from_antlr (node, parent)
			when C_kw_default then
				Result := create {AST_SWITCH_DEFAULT}.make_from_antlr (node, parent)
			end
		ensure
			Result /= Void and then Result.parent = parent
		end

	process_expression (node: ANTLR_AST_NODE; parent: AST_NODE): AST_EXPRESSION
		require
			node /= Void
		do
			inspect
				node.type
			when Ast_expression then
				Result := process_expression (node.children.first, parent)
			when Ast_function_call then
				Result := create {AST_FUNCTION_CALL}.make_from_antlr (node, parent)
			when Ast_builtin_va_arg_call then
				Result := create {AST_BUILTIN_VA_ARG_CALL}.make_from_antlr (node, parent)
			when Ast_alignof_call then
				Result := create {AST_ALIGNOF_CALL}.make_from_antlr (node, parent)
			when Ast_postfix_expression then
				Result := create {AST_POSTFIX_EXPRESSION}.make_from_antlr (node, parent)
			when Ast_hex_constant then
				Result := create {AST_HEX_CONSTANT}.make_from_antlr (node, parent)
			when Ast_octal_constant then
				Result := create {AST_OCTAL_CONSTANT}.make_from_antlr (node, parent)
			when Ast_decimal_constant then
				Result := create {AST_DECIMAL_CONSTANT}.make_from_antlr (node, parent)
			when Ast_character_constant then
				Result := create {AST_CHARACTER_CONSTANT}.make_from_antlr (node, parent)
			when Ast_string_constant then
				Result := create {AST_STRING_CONSTANT}.make_from_antlr (node, parent)
			when Ast_floating_point_constant then
				Result := create {AST_FLOATING_POINT_CONSTANT}.make_from_antlr (node, parent)
			when Ast_parenthesized_expression then
				Result := create {AST_PARENTHESIZED_EXPRESSION}.make_from_antlr (node, parent)
			when Ast_variable_expression then
				Result := create {AST_VARIABLE_EXPRESSION}.make_from_antlr (node, parent)
			when Ast_cast_expression then
				Result := create {AST_CAST_EXPRESSION}.make_from_antlr (node, parent)
			when Ast_assignment_expression then
				Result := create {AST_ASSIGNMENT_EXPRESSION}.make_from_antlr (node, parent)
			when Ast_unary_expression then
				Result := create {AST_UNARY_EXPRESSION}.make_from_antlr (node, parent)
			when Ast_array_initializer then
				Result := create {AST_INITIALIZER}.make_from_antlr (node, parent)
			when C_kw_sizeof then
				Result := create {AST_SIZEOF_EXPRESSION}.make_from_antlr (node, parent)
			when
				C_op_smaller, C_op_greater, C_op_smaller_equal, C_op_greater_equal,
				C_op_mult, C_op_div, C_op_modulo,
				C_op_equal, C_op_not_equal,
				C_op_plus, C_op_minus,
				C_op_logical_and, C_op_logical_or,
				C_op_bitand, C_op_bitor, C_op_exor,
				C_op_left_shift, C_op_right_shift
			then
				Result := create {AST_GENERIC_EXPRESSION}.make_from_antlr (node, parent)
			end
		ensure
			Result /= Void and then Result.parent = parent
		end

	process_type (node: ANTLR_AST_NODE; parent: AST_NODE): AST_TYPE
		require
			node /= Void
		local
			l_h: AST_VARIABLE_DECLARATION_OR_TYPE_HELPER
		do
			inspect
				node.type
			when Ast_type then
				node.children.start
				Result := process_type (node.children.item, Void)
				node.children.forth
				if not node.children.after and then node.children.item.type = Ast_type_qualifier then
					across node.children.item.children as c loop
						inspect
							c.item.type
						when C_kw_const then    Result.is_const := True
						when C_kw_volatile then Result.is_volatile := True
						end
					end
					node.children.forth
				end
				if not node.children.after then
					Result := create {AST_POINTER_TYPE}.make_from_antlr (node.children.item, Result, parent)
				else
					Result.parent := parent
				end
			when Ast_basic_type then
				Result := create {AST_BASIC_TYPE}.make_from_antlr (node, parent)
			when C_kw_struct, C_kw_union then
				Result := create {AST_STRUCT_TYPE}.make_from_antlr (node, parent)
			when C_kw_enum then
				Result := create {AST_ENUM_TYPE}.make_from_antlr (node, parent)
			when Ast_typedef_type then
				Result := create {AST_TYPEDEF_TYPE}.make_from_antlr (node, parent)
			when Ast_variable_declaration_or_type then
				l_h := create {AST_VARIABLE_DECLARATION_OR_TYPE_HELPER}.make_from_antlr (node)
				if l_h.is_type then
					Result := l_h.create_type
					Result.parent := parent
				end
			end
		ensure
			Result /= Void and then Result.parent = parent
		end

end
