class
	AST_ARRAY_TYPE

inherit
	AST_TYPE
		redefine
			make_copy
		end

create
	make,
	make_from_antlr,
	make_copy

feature {NONE}

	make (a_type: AST_TYPE; a_dimension: AST_EXPRESSION)
		require
			a_type /= Void and then a_type.parent = Void
			a_dimension /= Void implies a_dimension.parent = Void
		do
			type := a_type
			type.parent := Current
			if a_dimension /= Void then
				dimension := a_dimension
				dimension.parent := Current
				orig_c_dimension := dimension.ast_twin
				orig_c_dimension.is_orig_c_expression := True
			end
		end

	make_from_antlr (node: ANTLR_AST_NODE; a_type: AST_TYPE; init_parent: AST_NODE)
		require
			node /= Void and then node.type = Ast_array_type
			a_type /= Void and then a_type.parent = Void
		local
			cur_type: AST_TYPE
			cur_dimension: AST_EXPRESSION
		do
			parent := init_parent
			from
				cur_type := a_type
				node.children.finish
			invariant
				node.children.before or else node.children.item.type = C_op_close_square_bracket
			until
				node.children.before
			loop
				node.children.back
				if node.children.item.type /= C_op_open_square_bracket then
					cur_dimension := process_expression (node.children.item, Void)
					node.children.back
				else
					cur_dimension := Void
				end
				if node.children.index = 1 then
					make (cur_type, cur_dimension)
				else
					cur_type := create {AST_ARRAY_TYPE}.make (cur_type, cur_dimension)
				end
				node.children.back
			end
		end

feature {ANY}

	type: AST_TYPE

	dimension: AST_EXPRESSION

	orig_c_dimension: AST_EXPRESSION

	non_array_type: AST_TYPE
		do
			if attached {AST_ARRAY_TYPE} type as tmp then
				Result := tmp.non_array_type
			else
				Result := type
			end
		end

	array_count: INTEGER
		do
			if attached {AST_ARRAY_TYPE} type as tmp then
				Result := tmp.array_count + 1
			else
				Result := 1
			end
		end

	as_pointer: AST_POINTER_TYPE
		do
			if attached {AST_ARRAY_TYPE} type as tmp then
				Result := create {AST_POINTER_TYPE}.make (tmp.as_pointer)
			else
				Result := create {AST_POINTER_TYPE}.make (type.ast_twin)
			end
		end

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_array_type (Current)
			if visit_children then
				visit_node (type, v)
				visit_node (dimension, v)
			end
			v.end_visit_array_type (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (other: like Current)
		do
			Precursor (other)
			type := other.type.ast_twin
			type.parent := Current
			if other.dimension /= Void then
				dimension := other.dimension.ast_twin
				dimension.parent := Current
			end
			orig_c_dimension := other.orig_c_dimension -- won't be changed
		end

feature {ANY} -- AST_TYPE

	is_equal_type (a_type: AST_TYPE): BOOLEAN
		do
			if attached {AST_ARRAY_TYPE} a_type as other_type then
				-- TODO: also compare dimension?
				Result := type.is_equal_type (other_type.type)
			end
		end

invariant
	type /= Void
	(dimension = Void) = (orig_c_dimension = Void)

end
