class
	ENUM_TYPE_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			end_visit_enum_declaration,
			end_visit_enum_type
		end

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			create enum_declarations.make (16)
			transformer.tree.accept (Current)
			enum_declarations := Void
			Result := True
		end

	description: STRING
		do
			Result := "Transforming enum types"
		end

feature {NONE}

	enum_declarations: HASH_TABLE [AST_ENUM_DECLARATION, STRING_8]
			-- enum name -> enum declaration

feature {ANY} -- AST_DEFAULT_VISITOR

	end_visit_enum_declaration (node: AST_ENUM_DECLARATION)
		do
			if not node.is_forward_declaration then
				enum_declarations[node.identifier] := node
			end
		end

	end_visit_enum_type (node: AST_ENUM_TYPE)
		local
			l_enum_declaration: AST_ENUM_DECLARATION
			l_done: BOOLEAN
		do
			l_enum_declaration := enum_declarations[node.identifier]

			-- `(enum foo)3' -> `{FOO}.constant_name'
			if
				attached {AST_CAST_EXPRESSION} node.parent as cast_expression and then
				attached {AST_DECIMAL_CONSTANT} cast_expression.expression as decimal_constant
			then
				across
					l_enum_declaration.fields as c
				until
					l_done
				loop
					if
						attached {AST_DECIMAL_CONSTANT} c.item.initializer as tmp and then
						tmp.value.is_equal (decimal_constant.value)
					then
						replace_node (cast_expression, create {AST_EIFFEL_EXPRESSION}.make_mixed (
								<<
									"{", node.ast_twin, "}.",
									create {AST_VARIABLE_EXPRESSION}.make (c.item)
								>>,
								l_enum_declaration.replacement_type
							))
						l_done := True
					end
				end
			end

			-- `enum foo f;' -> `int f;'
			if not l_done then
				replace_node (node, l_enum_declaration.replacement_type.ast_twin)
			end
		end

end
