class
	LONGJMP_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			end_visit_function_call,
			end_visit_variable_declaration
		end

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			transformer.tree.accept (Current)
			across transformer.tree.declarations.twin as c loop
				if
					c.item.source_info.is_system_source and then
					c.item.source_info.source_file.is_equal ("setjmp.h")
				then
					delete_node (c.item)
				end
			end
			Result := True
		end

	description: STRING
		do
			Result := "Transforming setjmp/longjmp"
		end

feature {ANY} -- AST_DEFAULT_VISITOR

	end_visit_function_call (node: AST_FUNCTION_CALL)
		local
			l_cur: AST_NODE
		do
			-- `tmp = setjmp((struct __jmp_buf_tag *)(buf));'
			-- ->
			-- `buf := create {CE_EXCEPTION}.make ($routine_name, l_pos, $tmp)'
			-- `when X then -- continue here on raising buf:'
			if
				node.identifier_str /= Void and then
				(
					(trim (node.identifier_str).is_equal ("setjmp") and node.parameters.count = 1) or
					(trim (node.identifier_str).is_equal ("sigsetjmp") and node.parameters.count = 2)
				)
			then
				if
					attached {AST_ASSIGNMENT_EXPRESSION} node.parent as ae and then
					attached {AST_VARIABLE_EXPRESSION} ae.target as value_var and then
					attached {AST_FUNCTION_DECLARATION} value_var.declaration_binding.parent as fd and then -- can't take pointer ($) to global var (moving GC)
					attached {AST_EXPRESSION_STATEMENT} ae.parent as es and then
					attached get_variable_expression (node.parameters.first) as exception_var
				then
					replace_node (ae, create {AST_ASSIGNMENT_EXPRESSION}.make (
							exception_var.ast_twin,
							create {AST_EIFFEL_EXPRESSION}.make_mixed (
									<<"create {CE_EXCEPTION}.make ($" + fd.identifier + ", l_pos, $", value_var.ast_twin, ")">>,
									create {AST_BASIC_TYPE}.make (Ast_bt_void)
								)
						))

					insert_adjacent_statement (es, create {AST_EIFFEL_JUMP_TARGET}.make_comment (0, "continue here on raising " + exception_var.identifier + ":"), True)
					from
						l_cur := es.parent
					until
						l_cur = fd
					loop
						if attached {AST_IF_STATEMENT} l_cur as tmp then
							tmp.requires_split := True
						elseif attached {AST_WHILE_STATEMENT} l_cur as tmp then
							tmp.requires_split := True
						end
						l_cur := l_cur.parent
					end

					fd.is_longjmp_target := True
					fd.has_jump_statements := True
				else
					log ("Unsupported %"setjmp%" call: " + node.c_code)
				end

			-- `longjmp(buf, 1);' -> `buf.raise (1)'
			elseif
				node.identifier_str /= Void and then
				(
					trim (node.identifier_str).is_equal ("longjmp") or
					trim (node.identifier_str).is_equal ("siglongjmp")
				) and then
				node.parameters.count = 2
			then
				if
					attached get_variable_expression (node.parameters.first) as exception_var
				then
					replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (
							<<exception_var.ast_twin, ".raise (", node.parameters.last.ast_twin, ")">>,
							create {AST_BASIC_TYPE}.make (Ast_bt_void)
						))
				else
					log ("Unsupported %"longjmp%" call: " + node.c_code)
				end

			-- `memcpy(foo, bar, sizeof(...))' -> `foo := bar'
			elseif
				node.identifier_str /= Void and then
				node.identifier_str.is_equal ("memcpy") and then
				node.parameters.count = 3 and then
				attached get_variable_expression (node.parameters.first) as target_var and then
				attached get_variable_expression (node.parameters.i_th (2)) as source_var and then
				attached {AST_EIFFEL_TYPE} target_var.declaration_binding.type as et and then
				et.type_name.is_equal ("CE_EXCEPTION")
			then
				replace_node (node, create {AST_ASSIGNMENT_EXPRESSION}.make (target_var.ast_twin, source_var.ast_twin))
			end
		end

	end_visit_variable_declaration (node: AST_VARIABLE_DECLARATION)
		do
			if
				attached {AST_ARRAY_TYPE} node.type as at and then
				attached {AST_STRUCT_TYPE} at.type as st and then
				st.identifier.is_equal ("__jmp_buf_tag")
			then
				-- `CE_ARRAY [SD_JMP_BUF_TAG]' -> `CE_EXCEPTION'
				node.set_type (create {AST_EIFFEL_TYPE}.make ("CE_EXCEPTION"))
			end
		end

feature {NONE}

	trim (a_identifier: STRING_8): STRING_8
			-- trim preceding underscores
		require
			a_identifier /= Void
		local
			i: INTEGER_32
		do
			from
				i := 1
			until
				a_identifier[i] /= '_'
			loop
				i := i + 1
			end
			if i = 1 then
				Result := a_identifier
			else
				Result := a_identifier.substring (i, a_identifier.count)
			end
		ensure
			Result /= Void
		end

	get_variable_expression (node: AST_EXPRESSION): AST_VARIABLE_EXPRESSION
		require
			node /= Void
		local
			l_res: TUPLE [var: AST_VARIABLE_EXPRESSION]
			l_agent_visitor: AST_AGENT_VISITOR
		do
			create l_res
			create l_agent_visitor
			l_agent_visitor.agents.end_visit_variable_expression := agent (ag_node: AST_VARIABLE_EXPRESSION; ag_res: TUPLE [var: AST_VARIABLE_EXPRESSION])
				do
					ag_res.var := ag_node
				end (?, l_res)
			node.accept (l_agent_visitor)
			Result := l_res.var
		end

end
