class
	OO_CONTRACT_ADDER

inherit
	TRANSFORMATION_STEP

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			if not transformer.project_settings.disable_oo_reengineering then
				gcc_attribute_precondition_count := 0
				gcc_attribute_postcondition_count := 0
				argument_not_void_count := 0

				check_gcc_attributes
				check_pointer_argument_voidness

--				log ("Contract generation statistics:")
--				log ("- " + gcc_attribute_precondition_count.out + " gcc attribute preconditions")
--				log ("- " + gcc_attribute_postcondition_count.out + " gcc attribute postconditions")
--				log ("- " + argument_not_void_count.out + " argument not void")
--				log ("")

--				if (create {C2EIFFEL_GUI_INSTANCE}).instance.instance /= Void then
--					show_result_dialog
--				end
			end
			Result := True
		end

	description: STRING
		do
			Result := "Adding contracts"
		end

feature {NONE} -- statistics

	gcc_attribute_precondition_count,
	gcc_attribute_postcondition_count,
	argument_not_void_count: NATURAL_32

feature {NONE}

	check_gcc_attributes
		local
			l_evaluator: INTEGER_CONSTANT_EXPRESSION_EVALUATOR
		do
			-- http://gcc.gnu.org/onlinedocs/gcc/Function-Attributes.html
			across transformer.tree.declarations as declaration_cursor loop
				if
					attached {AST_FUNCTION_DECLARATION} declaration_cursor.item as fd and then
					attached fd.gcc_attribute_specifier as gcc_as
				then
					across gcc_as.attributes as attribute_cursor loop
						if attribute_cursor.item.identifier.is_equal ("__noreturn__") then
							fd.oo_postconditions.extend (create {AST_EIFFEL_EXPRESSION}.make ("False", create {AST_BASIC_TYPE}.make (Ast_bt_bool)))
							fd.oo_postconditions.last.parent := fd

							gcc_attribute_postcondition_count := gcc_attribute_postcondition_count + 1
						elseif attribute_cursor.item.identifier.is_equal ("__nonnull__") then
							if
								fd.identifier.is_equal ("memcpy") or
								fd.identifier.is_equal ("memmove")
							then
								-- loosen precondition to common usages:
								fd.oo_preconditions.extend (create {AST_EIFFEL_EXPRESSION}.make (
										"a___n > 0 implies (not a___dest.is_void and not a___src.is_void)",
										create {AST_BASIC_TYPE}.make (Ast_bt_bool)
									))
								fd.oo_preconditions.last.parent := fd

								gcc_attribute_precondition_count := gcc_attribute_precondition_count + 1
							elseif attribute_cursor.item.expressions.is_empty then
								across fd.parameter_list.parameters as parameter_cursor loop
									add_nonnull_check (fd, parameter_cursor.item, parameter_cursor.target_index)
								end
							else
								across attribute_cursor.item.expressions as argument_index_cursor loop
									create l_evaluator.make (argument_index_cursor.item)
									add_nonnull_check (fd, fd.parameter_list.parameters.i_th (l_evaluator.integer), l_evaluator.integer)
								end
							end
						end
					end
				end
			end
		end

	add_nonnull_check (a_function: AST_FUNCTION_DECLARATION; a_parameter: AST_NODE; a_parameter_index: INTEGER_32)
		require
			a_function /= Void and a_parameter /= Void
			a_parameter.parent.parent = a_function
			attached {AST_PARAMETER_LIST} a_parameter.parent as pl and then pl.parameters.index_of (a_parameter, 1) = a_parameter_index
		do
			if attached {AST_VARIABLE_DECLARATION} a_parameter as vd then
				if attached {AST_POINTER_TYPE} vd.type then
					a_function.oo_preconditions.extend (create {AST_EIFFEL_EXPRESSION}.make_mixed (<<
							"not ",
							create {AST_VARIABLE_EXPRESSION}.make (vd),
							".is_void"
						>>, create {AST_BASIC_TYPE}.make (Ast_bt_bool)))
					a_function.oo_preconditions.last.parent := a_function

					gcc_attribute_precondition_count := gcc_attribute_precondition_count + 1
				end
			elseif attached {AST_POINTER_TYPE} a_parameter as pt then
				a_function.oo_preconditions.extend (create {AST_EIFFEL_EXPRESSION}.make (
						"not a_arg" + a_parameter_index.out + ".is_void",
						create {AST_BASIC_TYPE}.make (Ast_bt_bool)
					))
				a_function.oo_preconditions.last.parent := a_function

				gcc_attribute_precondition_count := gcc_attribute_precondition_count + 1
			end
		end

feature {NONE}

	check_pointer_argument_voidness
		do
			across transformer.tree.declarations as declaration_cursor loop
				if attached {AST_FUNCTION_DECLARATION} declaration_cursor.item as fd and then fd.is_regular then
					across fd.parameter_list.parameters as parameter_cursor loop
						if
							attached {AST_VARIABLE_DECLARATION} parameter_cursor.item as vd and then
							attached {AST_POINTER_TYPE} vd.type and then
							is_dereferenced (fd, vd)
						then
							fd.oo_preconditions.extend (create {AST_EIFFEL_EXPRESSION}.make_mixed (<<
									"not ",
									create {AST_VARIABLE_EXPRESSION}.make (vd),
									".is_void"
								>>, create {AST_BASIC_TYPE}.make (Ast_bt_bool)))
							fd.oo_preconditions.last.parent := fd

							argument_not_void_count := argument_not_void_count + 1
						end
					end
				end
			end
		end

	is_dereferenced (a_function: AST_FUNCTION_DECLARATION; a_argument: AST_VARIABLE_DECLARATION): BOOLEAN
		require
			a_function /= Void
			a_argument /= Void
			a_argument.parent.parent = a_function
			attached {AST_POINTER_TYPE} a_argument.type
		local
			l_decls: HASH_SET [AST_VARIABLE_DECLARATION]
		do
			create l_decls.make (16)
			l_decls.extend (a_argument)
			Result := is_dereferenced_rec (l_decls, a_function.cfg, create {HASH_SET [CFG_BASIC_BLOCK]}.make (16)) = Res_dereferenced
		end

	Res_dereferenced,
	Res_not_dereferenced,
	Res_recursion_hit: INTEGER_8 = unique

	is_dereferenced_rec (a_decls: HASH_SET [AST_VARIABLE_DECLARATION]; a_block: CFG_BASIC_BLOCK; a_visited: HASH_SET [CFG_BASIC_BLOCK]): INTEGER_8
		require
			a_decls /= Void
			a_block /= Void
			a_visited /= Void
		local
			l_dereferenced: CELL [BOOLEAN]
			l_visitor: AST_AGENT_VISITOR
			l_has_res_dereferenced, l_has_res_not_dereferenced, l_has_res_recursion_hit: BOOLEAN
		do
			if not a_visited.has (a_block) then
				a_visited.extend (a_block)
				create l_dereferenced.put (False)

				create l_visitor
				l_visitor.agents.visit_assignment_expression := agent (ag_node: AST_ASSIGNMENT_EXPRESSION; ag_decls: HASH_SET [AST_VARIABLE_DECLARATION]): BOOLEAN
					local
						l_source_decl, l_target_decl: AST_VARIABLE_DECLARATION
					do
						if
							attached {AST_VARIABLE_EXPRESSION} ag_node.source as ve and then
							attached {AST_VARIABLE_DECLARATION} ve.declaration_binding as vd
						then
							l_source_decl := vd
						end
						if
							attached {AST_VARIABLE_EXPRESSION} ag_node.target as ve and then
							attached {AST_VARIABLE_DECLARATION} ve.declaration_binding as vd
						then
							l_target_decl := vd
						end

						if l_source_decl /= Void and then ag_decls.has (l_source_decl) then
							if l_target_decl /= Void then
								ag_decls.extend (l_target_decl) -- TODO: only in the block of the assignment, not all
							else
								-- expressions like `l_ctx.item.hints'; ignore (for now?)
							end
						elseif
							l_target_decl /= Void and then ag_decls.has (l_target_decl) and then
							not (
									attached {AST_GENERIC_EXPRESSION} ag_node.source as ge and then
									(
										(attached {AST_VARIABLE_EXPRESSION} ge.left.non_parentheses_expression as l and then l.declaration_binding = l_target_decl) or
										(attached {AST_VARIABLE_EXPRESSION} ge.right.non_parentheses_expression as r and then r.declaration_binding = l_target_decl)
									)
								)
						then
							ag_decls.remove (l_target_decl) -- TODO: only in the block of the assignment, not all
						else
							Result := True
						end
					end (?, a_decls)
				l_visitor.agents.end_visit_variable_expression := agent (ag_node: AST_VARIABLE_EXPRESSION; ag_decls: HASH_SET [AST_VARIABLE_DECLARATION]; ag_dereferenced: CELL [BOOLEAN])
					do
						if
							attached {AST_VARIABLE_DECLARATION} ag_node.declaration_binding as var_decl and then
							ag_decls.has (var_decl)
						then
							-- `p->...' | `p[...]'
							if
								attached {AST_POSTFIX_EXPRESSION} ag_node.non_parentheses_parent as pe and then
								(pe.type = C_op_arrow or pe.type = C_op_open_square_bracket)
							then
								ag_dereferenced.put (True)

							-- `*p'
							elseif
								attached {AST_UNARY_EXPRESSION} ag_node.non_parentheses_parent as ue and then
								ue.type = C_op_mult
							then
								ag_dereferenced.put (True)

							-- `eif_string (p)'
							elseif
								attached {AST_EIFFEL_EXPRESSION} ag_node.non_parentheses_parent as ee and then
								attached ee.previous_string (ag_node.parentheses_parent) as left_string and then
								left_string.is_equal ("eif_string (")
							then
								ag_dereferenced.put (True)

							-- `foo (p)' with `foo' requiring the argument to be not void
							elseif
								attached {AST_FUNCTION_CALL} ag_node.non_parentheses_parent as func_call and then
								attached func_call.declaration_binding as called_func and then
								attached func_call.parameters.index_of (ag_node.parentheses_parent, 1) as index and then
								index >= 1 and then -- in `identifier_expr' otherwise
								index <= called_func.parameter_list.parameters.count and then -- vararg otherwise
								attached {AST_VARIABLE_DECLARATION} called_func.parameter_list.parameters.i_th (index) as called_var_decl and then
								attached ("not " + called_var_decl.identifier + ".is_void") as precondition and then
								across called_func.oo_preconditions as c some c.item.eiffel_code.is_equal (precondition) end
							then
								ag_dereferenced.put (True)
							end
						end
					end (?, a_decls, l_dereferenced)
				across
					a_block.nodes as c
				until
					l_dereferenced.item
				loop
					c.item.accept (l_visitor)
				end

				if l_dereferenced.item then
					Result := Res_dereferenced
				else
					across
						a_block.out_edges as c
					until
						l_has_res_not_dereferenced
					loop
						inspect is_dereferenced_rec (a_decls, c.item.to_block, a_visited)
						when Res_dereferenced then     l_has_res_dereferenced := True
						when Res_not_dereferenced then l_has_res_not_dereferenced := True
						when Res_recursion_hit then    l_has_res_recursion_hit := True
						end
					end
					    if l_has_res_not_dereferenced then Result := Res_not_dereferenced
					elseif l_has_res_dereferenced then     Result := Res_dereferenced
					elseif l_has_res_recursion_hit then    Result := Res_recursion_hit
					else                                   Result := Res_not_dereferenced
					end
				end
			else
				Result := Res_recursion_hit
			end
		ensure
			Result = Res_dereferenced or Result = Res_not_dereferenced or Result = Res_recursion_hit
		end

feature {NONE}

	show_result_dialog
		local
			l_grid_dialog: GRID_DIALOG
			l_system_row, l_project_row: EV_GRID_ROW
			l_function_row, l_preconditions_row, l_postconditions_row: EV_GRID_ROW
		do
			create l_grid_dialog.make ("Contract Generation Results")
			l_grid_dialog.set_columns (<<
					"Routine / Contracts"
				>>)
			l_grid_dialog.grid.enable_single_row_selection
			l_grid_dialog.grid.enable_tree


			l_grid_dialog.grid.set_item (1, 1, create {EV_GRID_LABEL_ITEM}.make_with_text ("System"))
			l_grid_dialog.grid.set_item (1, 2, create {EV_GRID_LABEL_ITEM}.make_with_text ("Project"))
			l_system_row := l_grid_dialog.grid.row (1)
			l_project_row := l_grid_dialog.grid.row (2)

			across transformer.tree.declarations as c loop
				if
					attached {AST_FUNCTION_DECLARATION} c.item as fd and then
					(not fd.oo_preconditions.is_empty or not fd.oo_postconditions.is_empty)
				then
					if fd.source_info.is_project_source then
						l_function_row := add_row (l_project_row, fd.identifier)
					else
						l_function_row := add_row (l_system_row, fd.identifier)
					end
					if not fd.oo_preconditions.is_empty then
						l_preconditions_row := add_row (l_function_row, "require")
						across fd.oo_preconditions as c2 loop
							add_row (l_preconditions_row, c2.item.eiffel_code).do_nothing
						end
					end
					if not fd.oo_postconditions.is_empty then
						l_postconditions_row := add_row (l_function_row, "ensure")
						across fd.oo_postconditions as c2 loop
							add_row (l_postconditions_row, c2.item.eiffel_code).do_nothing
						end
					end
				end
			end

			if l_system_row.is_expandable then
				l_system_row.expand
			end
			if l_project_row.is_expandable then
				l_project_row.expand
			end

			l_grid_dialog.resize_columns_to_content
			l_grid_dialog.show
		end

	add_row (a_parent_row: EV_GRID_ROW; a_text: STRING_8): EV_GRID_ROW
		require
			a_parent_row /= Void
			a_text /= Void
		do
			a_parent_row.insert_subrow (a_parent_row.subrow_count + 1)
			Result := a_parent_row.subrow (a_parent_row.subrow_count)
			Result.set_item (1, create {EV_GRID_LABEL_ITEM}.make_with_text (a_text))
		end

end
