class
	OO_INHERITANCE_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			end_visit_eiffel_expression
		end

feature {ANY} -- TRANSFORMATION_STEP

	Heuristic_fields,
	Heuristic_cast: INTEGER_8 = unique

	process_impl: BOOLEAN
		do
			if not transformer.project_settings.disable_oo_reengineering then
				create inheritance.make (16)

				analyze_field_types
				analyze_casts
				apply_heuristics
--				print_statistics

--				if (create {C2EIFFEL_GUI_INSTANCE}).instance.instance /= Void then
--					show_result_dialog
--				end

				inheritance := Void
			end
			Result := True
		end

	description: STRING
		do
			Result := "Inferring O-O inheritance structure"
		end

feature {NONE}

	inheritance: HASH_TABLE [HASH_TABLE [HASH_SET [INTEGER_8], AST_STRUCT_DECLARATION], AST_STRUCT_DECLARATION]
		-- subclass -> superclass -> heuristics

	analyze_field_types
		do
			across transformer.tree.declarations as c loop
				if attached {AST_STRUCT_DECLARATION} c.item as sd then
					analyze_field_types_for (sd)
				end
			end
		end

	analyze_field_types_for (a_subclass: AST_STRUCT_DECLARATION)
		require
			a_subclass /= Void
		do
			across transformer.tree.declarations as c loop
				if
					attached {AST_STRUCT_DECLARATION} c.item as superclass and then
					a_subclass.fields.count > superclass.fields.count and then
					equal_fields (superclass, a_subclass)
				then
					ensure_entry_exists (a_subclass, superclass)
					(inheritance[a_subclass])[superclass].extend (Heuristic_fields)
				end
			end
		end

	analyze_casts
		do
			transformer.tree.accept (Current)
		end

	apply_heuristics
		local
			l_score: NATURAL_8 -- 0 - 100
		do
			across inheritance as subclass_cursor loop
				across subclass_cursor.item as superclass_cursor loop
					l_score := 0

					if superclass_cursor.item.has (Heuristic_fields) then
						l_score := l_score + 10
					end
					if superclass_cursor.item.has (Heuristic_cast) then
						if superclass_cursor.item.has (Heuristic_fields) then
							l_score := l_score + 50
						end
					end

					if l_score >= 50 then
						subclass_cursor.key.oo_superclasses.extend (superclass_cursor.key)
					end
				end
			end
		end

	print_statistics
		local
			l_project_name: STRING_8
			l_heuristic_fields_count,
			l_heuristic_cast_count: NATURAL_32
		do
			l_project_name := transformer.tree.source_file_name
			across inheritance as subclass_cursor loop
				across subclass_cursor.item as superclass_cursor loop
					across superclass_cursor.item as heuristic_cursor loop
						inspect heuristic_cursor.item
						when Heuristic_fields then
							l_heuristic_fields_count := l_heuristic_fields_count + 1
						when Heuristic_cast then
							l_heuristic_cast_count := l_heuristic_cast_count + 1
							log (l_project_name + "	" + subclass_cursor.key.identifier + "	" + superclass_cursor.key.identifier)
						end
					end
				end
			end

			log ("Inheritance inferring statistics:")
			log ("- " + l_heuristic_fields_count.out + " heuristic_fields")
			log ("- " + l_heuristic_cast_count.out + " heuristic_cast")
			log ("")
		end

	show_result_dialog
		local
			l_red: EV_COLOR
			l_grid_dialog: GRID_DIALOG
			l_row: INTEGER_32
			l_text: STRING_8
			l_grid_item: EV_GRID_LABEL_ITEM
		do
			create l_red.make_with_8_bit_rgb (255, 0, 0)

			create l_grid_dialog.make ("Inheritance Analysis Results")
			l_grid_dialog.set_columns (<<
					"Subclass / Superclass(es)",
					"Fields (0.1)",
					"Casts (0.5)",
					"Inherits (>= 0.5)"
				>>)
			l_grid_dialog.grid.enable_single_row_selection
			l_grid_dialog.grid.enable_tree

			l_row := 1
			across inheritance as subclass_cursor loop
				l_grid_dialog.grid.set_item (1, l_row, create {EV_GRID_LABEL_ITEM}.make_with_text (subclass_cursor.key.identifier))
				l_row := l_row + 1

				l_grid_dialog.grid.insert_new_rows_parented (subclass_cursor.item.count, l_row, l_grid_dialog.grid.row (l_row - 1))
				across subclass_cursor.item as superclass_cursor loop
					l_grid_dialog.grid.set_item (1, l_row, create {EV_GRID_LABEL_ITEM}.make_with_text (superclass_cursor.key.identifier))

					if superclass_cursor.item.has (Heuristic_fields) then
						l_text := "x"
					else
						l_text := " "
					end
					create l_grid_item.make_with_text (l_text)
					l_grid_item.align_text_center
					l_grid_dialog.grid.set_item (2, l_row, l_grid_item)

					if superclass_cursor.item.has (Heuristic_cast) then
						l_text := "x"
					else
						l_text := " "
					end
					create l_grid_item.make_with_text (l_text)
					l_grid_item.align_text_center
					l_grid_dialog.grid.set_item (3, l_row, l_grid_item)

					if subclass_cursor.key.oo_superclasses.has (superclass_cursor.key) then
						l_grid_dialog.grid.row (l_row).set_foreground_color (l_red)
						l_grid_dialog.grid.row (l_row).parent_row.set_foreground_color (l_red)
						l_text := "Yes"
					else
						l_text := "No"
					end
					create l_grid_item.make_with_text (l_text)
					l_grid_item.align_text_center
					l_grid_dialog.grid.set_item (4, l_row, l_grid_item)

					l_row := l_row + 1
				end
			end

			l_grid_dialog.resize_columns_to_content
			l_grid_dialog.show
		end

feature {ANY} -- AST_DEFAULT_VISITOR

	end_visit_eiffel_expression (node: AST_EIFFEL_EXPRESSION)
		local
			l_subclass, l_superclass: AST_STRUCT_DECLARATION
		do
			if
				node.mixed.count = 5 and then
				attached pointer_to_struct (node.mixed.i_th (2)) as dest_struct and then
				node.is_string (3, "}.make_cast (") and then
				attached pointer_to_struct (node.mixed.i_th (4)) as src_struct and then
				dest_struct.is_union = src_struct.is_union and then
				dest_struct.fields.count /= src_struct.fields.count
			then
				across
					transformer.tree.declarations as c
				until
					l_subclass /= Void and l_superclass /= Void
				loop
					if
						attached {AST_STRUCT_DECLARATION} c.item as sd and then
						(sd = dest_struct or sd = src_struct)
					then
						if
							sd.fields.count > dest_struct.fields.count or
							sd.fields.count > src_struct.fields.count
						then
							l_subclass := sd
						else
							l_superclass := sd
						end
					end
				end
				ensure_entry_exists (l_subclass, l_superclass)
				(inheritance[l_subclass])[l_superclass].extend (Heuristic_cast)
			end
		end

feature {NONE} -- utilities

	pointer_to_struct (a_node: ANY): AST_STRUCT_DECLARATION
		require
			a_node /= Void
		do
			if attached {AST_EXPRESSION} a_node as e then
				Result := pointer_to_struct (e.type_binding)
			elseif
				attached {AST_POINTER_TYPE} a_node as pt and then
				attached {AST_STRUCT_TYPE} pt.type as st
			then
				Result := st.declaration_binding
			elseif
				attached {AST_ARRAY_TYPE} a_node as at and then
				attached {AST_STRUCT_TYPE} at.type as st
			then
				Result := st.declaration_binding
			end
		end

	equal_fields (a_class1, a_class2: AST_STRUCT_DECLARATION): BOOLEAN
		require
			a_class1 /= Void
			a_class2 /= Void
		local
			l_fields1, l_fields2: LIST [AST_STRUCT_FIELD]
			l_struct1_cursor, l_struct2_cursor: CURSOR
		do
			l_fields1 := a_class1.fields
			l_fields2 := a_class2.fields
			l_struct1_cursor := l_fields1.cursor
			l_struct2_cursor := l_fields2.cursor
			from
				Result := True
				l_fields1.start
				l_fields2.start
			until
				Result = False or
				l_fields1.after or
				l_fields2.after
			loop
				Result := l_fields1.item.type.is_equal_type (l_fields2.item.type)
				l_fields1.forth
				l_fields2.forth
			end
			l_fields1.go_to (l_struct1_cursor)
			l_fields2.go_to (l_struct2_cursor)
		end

	ensure_entry_exists (a_subclass, a_superclass: AST_STRUCT_DECLARATION)
		require
			a_subclass /= Void
			a_superclass /= Void
		do
			if not inheritance.has (a_subclass) then
				inheritance[a_subclass] := create {HASH_TABLE [HASH_SET [INTEGER_8], AST_STRUCT_DECLARATION]}.make (16)
			end
			if not inheritance[a_subclass].has (a_superclass) then
				(inheritance[a_subclass])[a_superclass] := create {HASH_SET [INTEGER_8]}.make (16)
			end
		end

end
