class
	SIZEOF_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			end_visit_sizeof_expression
		end
	NAME_UTILITIES

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			transformer.tree.accept (Current)
			Result := True
		end

	description: STRING
		do
			Result := "Transforming %"sizeof()%" expressions"
		end

feature {ANY} -- AST_DEFAULT_VISITOR

	end_visit_sizeof_expression (node: AST_SIZEOF_EXPRESSION)
		do
			if attached handle (node.node) as expr then
				replace_node (node, expr)
			else
				log ("Unsupported sizeof expression: " + node.c_code)
			end
		end

feature {NONE}

	handle (node: AST_NODE): AST_EXPRESSION
		require
			node /= Void
		local
			name: STRING_8
		do
			-- In C, `sizeof()' has type `unsigned int'. In Eiffel on the other hand `int' is used.
			-- This means that CIL didn't give us the required casts. So we also adapt it here.
			if attached {AST_BASIC_TYPE} node as bt then
				if bt.type = Ast_bt_void then
					-- `sizeof(void)'  ->  `(1).to_natural_32'
					Result := create {AST_EIFFEL_EXPRESSION}.make (
							"(1).to_natural_32",
							create {AST_BASIC_TYPE}.make (Ast_bt_unsigned_int)
						)
				elseif bt.type = Ast_bt_builtin_va_list then
					-- `sizeof(va_list)' -> `{PLATFORM}.Pointer_bytes.to_natural_32'
					Result := create {AST_EIFFEL_EXPRESSION}.make (
							"{PLATFORM}.Pointer_bytes.to_natural_32",
							create {AST_BASIC_TYPE}.make (Ast_bt_unsigned_int)
						)
				else
					-- `sizeof(int)'  ->  `{PLATFORM}.Integer_32_bytes.to_natural_32'
					name := bt.eiffel_code.as_lower
					name[1] := name[1].as_upper
					if bt.type = Ast_bt_long_double then
						name := "{CE_UTILITIES}." + name
					else
						name := "{PLATFORM}." + name
					end
					Result := create {AST_EIFFEL_EXPRESSION}.make (
							name + "_bytes.to_natural_32",
							create {AST_BASIC_TYPE}.make (Ast_bt_unsigned_int)
						)
				end
			elseif attached {AST_STRUCT_TYPE} node as st then
				-- `sizeof(struct _foo)'  ->  `(create {FOO}.make).structure_size.to_natural_32'
				Result := create {AST_EIFFEL_EXPRESSION}.make_mixed (
						<<"(create {", st.ast_twin, "}.make).structure_size.to_natural_32">>,
						create {AST_BASIC_TYPE}.make (Ast_bt_unsigned_int)
					)
			elseif attached {AST_POINTER_TYPE} node then
				-- `sizeof(void *)' -> `{PLATFORM}.Pointer_bytes.to_natural_32'
				Result := create {AST_EIFFEL_EXPRESSION}.make (
						"{PLATFORM}.Pointer_bytes.to_natural_32",
						create {AST_BASIC_TYPE}.make (Ast_bt_unsigned_int)
					)
			elseif attached {AST_ARRAY_TYPE} node as at and then at.dimension /= Void then
				-- `sizeof(int [5])' -> `({PLATFORM}.Integer_32_bytes.to_natural_32 * (unsigned int)5)'
				Result := create {AST_PARENTHESIZED_EXPRESSION}.make (
						create {AST_GENERIC_EXPRESSION}.make (
								C_op_mult,
								handle (at.type),
								create {AST_CAST_EXPRESSION}.make (create {AST_BASIC_TYPE}.make (Ast_bt_unsigned_int), at.dimension.ast_twin)
							)
					)
			elseif attached {AST_STRING_CONSTANT} node as sc then
				-- `sizeof("abc")' -> `(("abc").count + 1).to_natural_32'
				sc.is_eiffel_string := True
				Result := create {AST_EIFFEL_EXPRESSION}.make_mixed (
						<<"((", sc.ast_twin, ").count + 1).to_natural_32">>,
						create {AST_BASIC_TYPE}.make (Ast_bt_unsigned_int)
					)
			elseif attached {AST_VARIABLE_EXPRESSION} node as ve then
				Result := handle (ve.type_binding)
			elseif attached {AST_EXPRESSION} node as e then
				-- Don't work with the expression itself. `sizeof(foo->bar)' for instance can't be `foo.item.bar.sizeof'
				-- because `foo' might point to nothing. Translate it to e.g. `Pointer_bytes'.
				Result := handle (e.type_binding)
			end
		end

end
