class
	NATIVE_FUNCTIONS_WRITER

inherit
	NAME_UTILITIES

feature {ANY}

	write_native_variable (node: AST_VARIABLE_DECLARATION; buffer: STRING)
		require
			node /= Void and then node.is_extern
			buffer /= Void
		do
			-- 1. getter
			if needs_wrapping (node) then
				-- foo: CE_POINTER [INTEGER_8] assign set_foo
				--    do
				--       create Result.make_shared (c_foo)
				--    end
				buffer.append (node.identifier + ": " + node.type.eiffel_code)
				if
					not node.type.is_const and
					not node.changed_to_array
				then
					buffer.append (" assign set_" + node.identifier)
				end
				buffer.append ("%N")
				buffer.append ("		do%N")
				buffer.append ("			create ")
				if
					(attached {AST_POINTER_TYPE} node.type as pt and then not attached {AST_FUNCTION_TYPE} pt.type) or
					attached {AST_ARRAY_TYPE} node.type
				then
					buffer.append ("{" + effective_type (node.type) + "} ")
				end
				if attached {AST_ARRAY_TYPE} node.type as at then
					if at.dimension /= Void then
						buffer.append ("Result.make_shared (c_" + node.identifier + ", <<" + at.dimension.eiffel_code + ">>)%N")
					else
						buffer.append ("Result.make_shared_from_pointer (c_" + node.identifier + ")%N")
					end
				else
					buffer.append ("Result.make_shared (c_" + node.identifier + ")%N")
				end
				buffer.append ("		end%N%N")

				buffer.append ("	c_")
			end

			buffer.append (node.identifier + ": ")
			if needs_wrapping (node) then
				buffer.append ("POINTER")
			else
				buffer.append (node.type.eiffel_code)
				if not node.type.is_const then
					buffer.append (" assign set_" + node.identifier)
				end
			end
			buffer.append ("%N")
			buffer.append ("		external%N")
			buffer.append ("			%"C inline use <" + c_header_file ({AST_TRANSLATION_UNIT} #? node.parent) + ">%"%N")
			buffer.append ("		alias%N")
			buffer.append ("			%"")

			buffer.append ("return ")
			if needs_wrapping (node) then
				buffer.append ("(EIF_POINTER)")
			end
			if
				attached {AST_STRUCT_TYPE} node.type or
				node.changed_to_array
			then
				buffer.append ("&")
			end
			buffer.append (node.orig_c_identifier)
			buffer.append (";%"%N")
			buffer.append ("		end")

			-- 2. setter
			if
				not node.type.is_const and
				not node.changed_to_array
			then
				buffer.append ("%N%N")

				if needs_wrapping (node) then
					-- set_foo (a_foo: CE_POINTER [INTEGER_8])
					--    do
					--       c_set_foo (a_foo.c_pointer)
					--    end

					buffer.append ("	set_" + node.identifier + " (a_" + node.identifier + ": " + node.type.eiffel_code + ")%N")
					buffer.append ("		do%N")
					buffer.append ("			c_set_" + node.identifier + " (a_" + node.identifier + ".c_pointer)%N")
					buffer.append ("		end%N%N")
					buffer.append ("	c_")
				else
					buffer.append ("	")
				end

				buffer.append ("set_" + node.identifier + " (a_" + node.identifier + ": ")
				if needs_wrapping (node) then
					buffer.append ("POINTER")
				else
					buffer.append (node.type.eiffel_code)
				end
				buffer.append (")%N")
				buffer.append ("		external%N")
				buffer.append ("			%"C inline use <" + c_header_file ({AST_TRANSLATION_UNIT} #? node.parent) + ">%"%N")
				buffer.append ("		alias%N")
				buffer.append ("			%"")

				if attached {AST_BASIC_TYPE} node.type then
					-- ospeed = $a_ospeed;
					buffer.append (node.orig_c_identifier + " = $a_" + node.identifier.as_lower)
				elseif attached {AST_STRUCT_TYPE} node.type as tmp then
					-- widgetClassRec = *(struct _WidgetClassRec *)$a_widgetclassrec;
					buffer.append (node.orig_c_identifier + " = *(" + tmp.c_code + " *)$a_" + node.identifier.as_lower)
				elseif attached {AST_POINTER_TYPE} node.type as tmp then
					-- foo = (bar *)$a_foo;
					buffer.append (node.orig_c_identifier + " = (" + tmp.c_code + ")$a_" + node.identifier.as_lower)
				else
					buffer.append ("printf(%%%"NATIVE_FUNCTIONS_WRITER.write_native_variable: " + node.type.generator + "\n%%%")")
				end

				buffer.append (";%"%N")
				buffer.append ("		end")
			end
		end

	write_native_function (node: AST_FUNCTION_DECLARATION; buffer: STRING)
		require
			node /= Void and then node.is_extern
			buffer /= Void
		local
			l_has_return, l_needs_wrapper: BOOLEAN
		do
			if node.orig_c_identifier.is_equal ("malloc") then
				buffer.append ("	malloc (a_size: NATURAL_32): CE_POINTER [ANY]%N")
				buffer.append ("		do%N")
				buffer.append ("			create {CE_VOID_POINTER} Result.make_shared_with_count (%N")
				buffer.append ("					default_pointer.memory_alloc (a_size.to_integer_32),%N")
				buffer.append ("					a_size.to_integer_32%N")
				buffer.append ("				)%N")
				buffer.append ("		end")
			elseif node.orig_c_identifier.is_equal ("calloc") then
				buffer.append ("	calloc (a_nmemb: NATURAL_32; a_size: NATURAL_32): CE_POINTER [ANY]%N")
				buffer.append ("		do%N")
				buffer.append ("			create {CE_VOID_POINTER} Result.make_shared_with_count (%N")
				buffer.append ("					default_pointer.memory_calloc (a_nmemb.to_integer_32, a_size.to_integer_32),%N")
				buffer.append ("					(a_nmemb * a_size).to_integer_32%N")
				buffer.append ("				)%N")
				buffer.append ("		end")
			elseif node.orig_c_identifier.is_equal ("realloc") then
				buffer.append ("	realloc (a_ptr: CE_POINTER [ANY]; a_size: NATURAL_32): CE_POINTER [ANY]%N")
				buffer.append ("		do%N")
				buffer.append ("			create {CE_VOID_POINTER} Result.make_shared_with_count (%N")
				buffer.append ("					a_ptr.c_pointer.memory_realloc (a_size.to_integer_32),%N")
				buffer.append ("					a_size.to_integer_32%N")
				buffer.append ("				)%N")
				buffer.append ("		end")
			else
				l_has_return := not attached {AST_BASIC_TYPE} node.return_type as basic_type or else basic_type.type /= Ast_bt_void
				l_needs_wrapper := node.parameter_list.is_vararg or
						needs_wrapping (node.return_type) or
						across node.parameter_list.parameters as c some needs_wrapping (c.item) end

				if l_needs_wrapper then
					-- strchr (a_s: CE_POINTER [INTEGER_8]; a_c: INTEGER_32): CE_POINTER [INTEGER_8]
					--    do
					--       create Result.make_shared (c_strchr (a_s.c_pointer, a_c))
					--    end

					buffer.append ("	" + node.identifier)
					if not node.parameter_list.parameters.is_empty then
						buffer.append (" (")
						buffer.append (node.parameter_list.eiffel_code)
						buffer.append (")")
					end
					if l_has_return then
						buffer.append (": ")
						buffer.append (node.return_type.eiffel_code)
					end
					buffer.append ("%N")

					if not node.oo_preconditions.is_empty then
						buffer.append ("		require%N")
						across node.oo_preconditions as c loop
							buffer.append ("			" + c.item.eiffel_code + "%N")
						end
					end

					buffer.append ("		do%N")

					buffer.append ("			")
					if l_has_return then
						if needs_wrapping (node.return_type) then
							if
								(attached {AST_POINTER_TYPE} node.return_type as pt and then not attached {AST_FUNCTION_TYPE} pt.type) or
								attached {AST_ARRAY_TYPE} node.return_type
							then
								buffer.append ("create {" + effective_type (node.return_type) + "} Result.make_shared (")
							else
								buffer.append ("create Result.make_shared (")
							end
						else
							buffer.append ("Result := ")
						end
					end
					if node.parameter_list.is_vararg then
						buffer.append ("(create {CE_ROUTINE}.make_shared (c_" + node.identifier + ")).")
						if not l_has_return then
							buffer.append ("call")
						elseif attached {AST_BASIC_TYPE} node.return_type as bt then
							buffer.append (bt.eiffel_code.as_lower + "_item")
						elseif attached {AST_POINTER_TYPE} node.return_type as pt then
							buffer.append ("pointer_item")
						end
						buffer.append (" (a_args)")
						if attached {AST_POINTER_TYPE} node.return_type as pt then
							buffer.append (".c_pointer")
						end
					else
						buffer.append ("c_" + node.identifier)
						if not node.parameter_list.parameters.is_empty then
							buffer.append (" (")
							if node.parameter_list.is_vararg then
								buffer.append ("a_args")
							else
								across node.parameter_list.parameters as c loop
									if c.target_index > 1 then
										buffer.append (", ")
									end
									buffer.append (argument_name (c))
									if needs_wrapping (c.item) then
										buffer.append (".c_pointer")
									end
								end
							end
							buffer.append (")")
						end
					end
					if l_has_return and needs_wrapping (node.return_type) then
						buffer.append (")")
					end
					buffer.append ("%N")

					if not node.oo_postconditions.is_empty then
						buffer.append ("		ensure%N")
						across node.oo_postconditions as c loop
							buffer.append ("			" + c.item.eiffel_code + "%N")
						end
					end

					buffer.append ("		end%N%N")
				end

				-- c_strchr (a_s: POINTER; a_c: INTEGER_32): POINTER
				buffer.append ("	")
				if l_needs_wrapper then
					buffer.append ("c_")
				end
				buffer.append (node.identifier)
				if
					not node.parameter_list.is_vararg and
					not node.parameter_list.parameters.is_empty
				then
					buffer.append (" (")
					if node.parameter_list.is_vararg then
						buffer.append (node.parameter_list.eiffel_code)
					else
						across node.parameter_list.parameters as c loop
							if c.target_index > 1 then
								buffer.append ("; ")
							end
							buffer.append (argument_name (c))
							buffer.append (": ")
							if needs_wrapping (c.item) then
								buffer.append ("POINTER")
							else
								buffer.append (argument_type (c.item).eiffel_code)
							end
						end
					end
					buffer.append (")")
				end
				if node.parameter_list.is_vararg or l_has_return then
					buffer.append (": ")
					if node.parameter_list.is_vararg or needs_wrapping (node.return_type) then
						buffer.append ("POINTER")
					else
						buffer.append (node.return_type.eiffel_code)
					end
				end
				buffer.append ("%N")

				if
					not l_needs_wrapper and
					not node.oo_preconditions.is_empty
				then
					buffer.append ("		require%N")
					across node.oo_preconditions as c loop
						buffer.append ("			" + c.item.eiffel_code + "%N")
					end
				end

				-- external
				--    "C inline use <include/xeyes.h>"
				buffer.append ("		external%N")
				buffer.append ("			%"C inline use <" + c_header_file ({AST_TRANSLATION_UNIT} #? node.parent) + ">%"%N")

				-- alias
				buffer.append ("		alias%N")
				buffer.append ("			%"")
				if node.parameter_list.is_vararg then
					buffer.append ("return (EIF_POINTER)&" + node.orig_c_identifier)
				else
					-- return strchr($a_s, $a_c)
					if l_has_return then
						buffer.append ("return ")
						if needs_wrapping (node.return_type) then
							buffer.append ("(EIF_POINTER)")
						end
					end
					if
						node.orig_c_identifier.is_equal ("fork") and
						(attached {AST_BASIC_TYPE} node.return_type as bt and then bt.type = Ast_bt_int) and
						node.parameter_list.parameters.is_empty
					then
						buffer.append ("eiffel_fork")
					else
						buffer.append (node.orig_c_identifier)
					end
					buffer.append ("(")
					across node.parameter_list.parameters as c loop
						if c.target_index > 1 then
							buffer.append (", ")
						end
						if attached {AST_BASIC_TYPE} argument_type (c.item) as bt and then bt.type = Ast_bt_long_double then
							buffer.append ("*(long double *)")
						end
						buffer.append ("$")
						buffer.append (argument_name (c).as_lower)
					end
					buffer.append (")")
				end
				buffer.append (";%"%N")

				if
					not l_needs_wrapper and
					not node.oo_postconditions.is_empty
				then
					buffer.append ("		ensure%N")
					across node.oo_postconditions as c loop
						buffer.append ("			" + c.item.eiffel_code + "%N")
					end
				end

				buffer.append ("		end")
			end
		end

feature {NONE}

	needs_wrapping (a_node: AST_NODE): BOOLEAN
		require
			a_node /= Void
		do
			Result :=
				not attached {AST_BASIC_TYPE} argument_type (a_node) as bt or else
				(bt.type = Ast_bt_builtin_va_list or bt.type = Ast_bt_long_double)
		end

	argument_name (a_c: INDEXABLE_ITERATION_CURSOR [AST_NODE]): STRING_8
		require
			a_c /= Void
		do
			if attached {AST_VARIABLE_DECLARATION} a_c.item as tmp then
				Result := tmp.identifier
			elseif attached {AST_TYPE} a_c.item as tmp then
				Result := "a_arg" + a_c.target_index.out
			end
		ensure
			Result /= Void
		end

	argument_type (a_node: AST_NODE): AST_TYPE
		require
			attached {AST_VARIABLE_DECLARATION} a_node or attached {AST_TYPE} a_node
		do
			if attached {AST_VARIABLE_DECLARATION} a_node as tmp then
				Result := tmp.type
			elseif attached {AST_TYPE} a_node as tmp then
				Result := tmp
			end
		ensure
			Result /= Void
		end

end
